/*
 * Copyright (c) 2004-2005 Atmark Techno, Inc.  All Rights Reserved.
 */

#include <ep93xx/ioregs.h>
#include <ep93xx/memcpy.h>
#include <target/herrno.h>
#include <target/io.h>
#include <target/scan.h>
#include <target/str.h>
#include <target/buffer.h>
#include <target/memcmp.h>
#include "tags.h"
#include "gunzip.h"
#include "linux.h"
#include "memmap.h"
#include "memzero.h"
#include "flash.h"
#include "eth.h"
#include "board.h"

/* for setting the root device */
#define MINORBITS       8
#define MKDEV(ma,mi)    (((ma) << MINORBITS) | (mi))

#define RAMDISK_MAJOR   1	/* major for "RAM disk" */
#define RAMDISK0_MINOR	0	/* b 1 0 == /dev/ram0 */

#define ROM_FLASH_MAJOR	31	/* major for "ROM/flash memory card" */
#define FLASH0_MINOR	16	/* b 31 16 == /dev/flash0 */
#define FLASH1_MINOR	17	/* b 31 17 == /dev/flash1 */

static int major = RAMDISK_MAJOR;
static int minor = RAMDISK0_MINOR;

#define COMMAND_LINE_SIZE 1024

extern unsigned long bytes_out;

#if 0
#define PBDR_FLASHBOOT  0x80
#endif

#define SECTOR_SIZE     0x200
#define INODE_SIZE      0x80

static int linux_cmdfunc(int argc, char *argv[])
{
	struct tag *tag = (struct tag *) LINUX_PARAM_ADDRESS;

	/* zero param block */
	memzero (tag, LINUX_PARAM_SIZE);

	/* set up core tag */
	tag->hdr.tag = ATAG_CORE;
	tag->hdr.size = tag_size(tag_core);
	tag->u.core.flags = 0;
	tag->u.core.pagesize = 0x1000;
	tag->u.core.rootdev = MKDEV(major, minor);

	/* 8 MB of SDRAM at 0xc0000000 */
	tag = tag_next(tag);
	tag->hdr.tag = ATAG_MEM;
	tag->hdr.size = tag_size(tag_mem32);
	tag->u.mem.size = DRAM1_SIZE;
	tag->u.mem.start = DRAM1_START;

	/* 8 MB of SDRAM at 0xc1000000 */
	tag = tag_next(tag);
	tag->hdr.tag = ATAG_MEM;
	tag->hdr.size = tag_size(tag_mem32);
	tag->u.mem.size = DRAM2_SIZE;
	tag->u.mem.start = DRAM2_START;

	/* 16 MB of SDRAM at 0xc4800000 */
	tag = tag_next(tag);
	tag->hdr.tag = ATAG_MEM;
	tag->hdr.size = tag_size(tag_mem32);
	tag->u.mem.size = DRAM3_SIZE;
	tag->u.mem.start = DRAM3_START;

	if (major == RAMDISK_MAJOR) {
	  /* an initial ramdisk image in flash at 0x00700000 */
	  tag = tag_next(tag);
	  tag->hdr.tag = ATAG_INITRD;
	  tag->hdr.size = tag_size(tag_initrd);
	  tag->u.initrd.start = INITRD_LOAD_ADDRESS;
	  tag->u.initrd.size  = bytes_out;
	}

	/* the command line arguments */
	if (argc > 1) {
		tag = tag_next(tag);
		tag->hdr.tag = ATAG_CMDLINE;
		tag->hdr.size = (COMMAND_LINE_SIZE + 3 +
			 sizeof(struct tag_header)) >> 2;

		{
			const unsigned char *src;
			unsigned char *dst;
			dst = tag->u.cmdline.cmdline;
			memzero (dst, COMMAND_LINE_SIZE);
			while (--argc > 0) {
				src = *++argv;
				hprintf ("Doing %s\n", src);
				while (*src)
					*dst++ = *src++;
				*dst++ = ' ';
			}
			*--dst = '\0';
		}
	}

	tag = tag_next(tag);
	tag->hdr.tag = 0;
	tag->hdr.size = 0;

	/* branch to kernel image */
	__asm__ volatile (
	"	mov	r4, #0xC0000000\n"	/* start of DRAM */
	"	add	r4, r4, #0x00018000\n"	/* kernel offset */
	"	mov	r0, #0\n"		/* kernel sanity check */
	"	mov	r1, #0x100\n"           /* armadillo210 arch. number */
	"	orr     r1, r1, #0x82\n"	/* number = 0x320 (800) */
	"	mov	r2, #0\n"
	"	mov	r3, #0\n"
	"	mov	pc, r4"			/* go there! */
	);

	/* never get here */
	return 0;
}

static int unzip_cmdfunc(int argc, char *argv[])
{
	addr_t kernel;
	addr_t ramdisk;

	if (argc < 2 || argc > 3)
		return -H_EUSAGE;

	argv++;
	if (scan(*argv, &kernel))
		return -H_EADDR;
	
	argv++;
	if (scan(*argv, &ramdisk))
		return -H_EADDR;
	
	/* decompress kernel image to DRAM */
	gunzip_object (" kernel", kernel, LINUX_LOAD_ADDRESS);
	gunzip_object ("ramdisk", ramdisk, INITRD_LOAD_ADDRESS);

	return 0;
}

char mtdparts[] = { MTDPARTS };
char console[] = { "console=ttyAM1,115200" };

static void boost_on (void)
{
    const unsigned int offset = LINUX_LOAD_ADDRESS - 0x4000;
    unsigned int i;
    for (i = 0x0; i < 0x2000; i += 0x4) {
      *(volatile unsigned int *)(offset + i) = (i << 18) + 0xc1e;
    }
    for (       ; i < 0x2400; i += 0x4) {
      *(volatile unsigned int *)(offset + i) = (i << 18) + 0xc12;
    }
    for (       ; i < 0x4000; i += 0x4) {
      *(volatile unsigned int *)(offset + i) = (i << 18) + 0xc1e;
    }
    __asm__ volatile (
		      "    mov     r0, #0                  \n"
		      "    mcr     p15, 0, r0, c7, c7      \n"
		      "    mcr     p15, 0, r0, c7, c10, 4  \n"
		      "    mcr     p15, 0, r0, c8, c7      \n"
		      "    mov     r0, #0xc0000000         \n"
		      "    add     r0, r0, #0x14000        \n"
		      "    mcr     p15, 0, r0, c2, c0      \n"
		      "    mov     r0, #0x1f               \n"
		      "    mcr     p15, 0, r0, c3, c0      \n"
		      "    mrc     p15, 0, r0, c1, c0      \n"
		      "    orr     r0, r0, #0x0035         \n"
		      "    orr     r0, r0, #0x3100         \n"
		      "    mcr     p15, 0, r0, c1, c0      \n"
		      );
}

static void boost_off (void)
{
    __asm__ volatile (
		      "    mov     r1, #7 << 5             \n"
                      "1:  orr     r0, r1, #63 << 26       \n"
		      "2:  mcr     p15, 0, r0, c7, c14, 2  \n"
		      "    subs    r0, r0, #1 << 26        \n"
		      "    bcs     2b                      \n"
		      "    subs    r1, r1, #1 << 5         \n"
                      "    bcs     1b                      \n"
                      "    mov     r0, #0                  \n"
		      "    mcr     p15, 0, r0, c7, c5, 0   \n"
		      "    mcr     p15, 0, r0, c7, c10, 4  \n"
		      "    mrc     p15, 0, r0, c1, c0      \n"
		      "    bic     r0, r0, #0x0035         \n"
		      "    bic     r0, r0, #0x3100         \n"
		      "    mcr     p15, 0, r0, c1, c0      \n"
		      );
}

static int option_count = 0;
static char *options = 0;
static char *options_next = 0;
static int get_option_count (void)
{
  unsigned long next_offset;
  char *p;
  int new_flag = 1;
  
  option_count = 0;
  
  for (options = (char *)FLASH_OPTION_START;
       options < (char *)(FLASH_OPTION_START + FLASH_OPTION_SIZE);
       options += next_offset) {
    next_offset = *(unsigned long *)options;
    if (next_offset == (unsigned long)0xffffffff) {
      options += sizeof(unsigned long);
      break;
    }
  }
  if (options >= (char *)(FLASH_OPTION_START + FLASH_OPTION_SIZE)) {
    options = (char *)FLASH_OPTION_START + sizeof(unsigned long);
    options_next = (char *)FLASH_OPTION_START + sizeof(unsigned long);
    return 1;
  }
  for (p = options;
       p < (char *)(FLASH_OPTION_START + FLASH_OPTION_SIZE);
       p++) {
    switch (*p) {
    case 0x0:
      new_flag = 1;
      break;
    case 0xff:
      if (!option_count) {
	options_next = p;
	return 1;
      }
      if ((int)p % sizeof(unsigned long)) {
	p += sizeof(unsigned long) - ((int)p % sizeof(unsigned long));
      }
      options_next = p + sizeof(unsigned long);
      return option_count + 1;
    default:
      if (new_flag) {
	option_count++;
	new_flag = 0;
      }
      break;
    }
  }
  
  option_count = 0;
  options = (char *)FLASH_OPTION_START + sizeof(unsigned long);
  options_next = (char *)FLASH_OPTION_START + sizeof(unsigned long);
  return 1;
}

static void get_options (char *argv[])
{
  char *p;
  int i = 0;
  int new_flag = 1;
  
  argv[i++] = 0;

  if (!option_count) {
    return;
  }
  
  for (p = options; *p != 0xff; p++) {
    switch (*p) {
    case 0x0:
      new_flag = 1;
      break;
    default:
      if (new_flag) {
	argv[i++] = p;
	new_flag = 0;
      }
      break;
    }
  }
  
  return;
}

static void hex2str(char str[], unsigned long hex)
{
  int i;
  char c;
  
  str[0] = '0';
  str[1] = 'x';
  for (i = 0; i < 8; i++) {
    c = (hex >> ((7 - i) << 2)) & 0xf;
    str[2 + i] = ((c <= 9) ? '0' : ('a' - 10)) + c;
  }
  str[10] = 0;
}

static int setenv_cmdfunc(int argc, char *argv[])
{
  int i;
  unsigned char *p = dlbuf;
  char addr[16], size[16];
  int c_argc;
  char *c_argv[3];
  
  for (i = 1; i < argc; i++) {
    strcpy (p, argv[i]);
    p += strlen(p) + 1;
  }
  for (; (int)p % 4; p++) {
    *p = 0xff;
  }
  
  get_option_count();
  if (argc <= 1) {
    int offset = 0;
    for (i = 0; i < option_count; i++) {
      hprintf("%d: %s\n", i + 1, options + offset);
      offset += strlen(options + offset) + 1;
    }
    return 0;
  }
  if (options_next + (p - dlbuf) >=
      (char *)(FLASH_OPTION_START + FLASH_OPTION_SIZE)) {
    hex2str (addr, FLASH_OPTION_START);
    c_argc = 2;
    c_argv[0] = 0;
    c_argv[1] = addr;
    (*flash_erase_command.func)(c_argc, c_argv);
    option_count = 0;
    options = (char *)FLASH_OPTION_START + sizeof(unsigned long);
    options_next = (char *)FLASH_OPTION_START + sizeof(unsigned long);
  }
  else {
    if (options_next != options) {
      unsigned long temp = *(unsigned long *)dlbuf;
      *(unsigned long *)dlbuf = (unsigned long)(options_next - options);
      hex2str (addr, (unsigned long)(options - sizeof(unsigned long)));
      hex2str (size, sizeof(unsigned long));
      c_argc = 3;
      c_argv[0] = 0;
      c_argv[1] = addr;
      c_argv[2] = size;
      (*flash_program_command.func)(c_argc, c_argv);
      *(unsigned long *)dlbuf = temp;
    }
  }
  
  hex2str (addr, (unsigned long)options_next);
  hex2str (size, p - dlbuf);
  c_argc = 3;
  c_argv[0] = 0;
  c_argv[1] = addr;
  c_argv[2] = size;
  (*flash_program_command.func)(c_argc, c_argv);
  
  return 0;
}

static int clearenv_cmdfunc(int argc, char *argv[])
{
  char addr[16];
  int c_argc;
  char *c_argv[3];
  
  hex2str (addr, FLASH_OPTION_START);
  c_argc = 2;
  c_argv[0] = 0;
  c_argv[1] = addr;
  (*flash_erase_command.func)(c_argc, c_argv);
  option_count = 0;
  options = (char *)FLASH_OPTION_START + sizeof(unsigned long);
  options_next = (char *)FLASH_OPTION_START + sizeof(unsigned long);
  
  return 0;
}

static int boot_cmdfunc(int argc, char *argv[])
{
	int i;
	int gunzip_kernel = 1;
	int o_argc = 0;
	char *o_argv[o_argc + 3];

	if(argc > 1){
	  o_argc = argc;
	}else{
	  o_argc = get_option_count();
	}
	
	if(argc > 1){
	  for(i = 0; i < argc; i++){
	    o_argv[i] = argv[i];
	  }
	}else{
	  if(o_argc <= 1){
	    o_argv[o_argc++] = console;
	  }else{
	    get_options(o_argv);
	  }
	}

	for(i=0; i<o_argc; i++){
	  if(strncmp(o_argv[i], "console=", 8) == 0){
	    if(strncmp(&o_argv[i][8], "ttyAM0", 6) == 0){
	      jumper_mode = 1;
	    }else if(strncmp(&o_argv[i][8], "ttyAM1", 6) == 0){
	      jumper_mode = 0;
	    }else{
	      jumper_mode = 0;
	    }
	    break;
	  }
	}

	major = RAMDISK_MAJOR;
	minor = RAMDISK0_MINOR;

	argc = o_argc;
	argv = o_argv;
	argv[argc++] = mtdparts;

	if(argc < 1)
		return -H_EUSAGE;
	
	enable_phy_module();
	LED_RED_ON();

	boost_on ();
	if(gunzip_kernel){
	  gunzip_object(" kernel", LINUX_SRC_ADDRESS, LINUX_LOAD_ADDRESS);
	}
	if(major == RAMDISK_MAJOR){
	  gunzip_object("ramdisk", INITRD_SRC_ADDRESS, INITRD_LOAD_ADDRESS);
	}
	boost_off();

	linux_cmdfunc(argc, argv);

	return 0;
}

const command_t linux_command =
	{ "linux", "<linux options>", "start Linux", &linux_cmdfunc };
const command_t unzip_command =
	{ "unzip", "<addr> <addr>", "unzip Linux kernel and ramdisk from addrs", &unzip_cmdfunc };
const command_t boot_command =
	{ "boot", "", "boot default Linux kernel and ramdisk", &boot_cmdfunc };
const command_t setenv_command =
	{ "setenv", "<linux options>", "set Linux boot options", &setenv_cmdfunc };
const command_t clearenv_command =
	{ "clearenv", "", "clear Linux boot options", &clearenv_cmdfunc };
