/*
 * Copyright (c) 2000 Blue Mug, Inc.  All Rights Reserved.
 */

#include <target/assert.h>
#include <target/crc.h>
#include <target/command.h>
#include <target/herrno.h>
#include <target/io.h>
#include <target/str.h>
#include <netarm/ioregs.h>
#include "bootmode.h"

typedef enum medium_type {
	MEDIUM_SERIAL = 0,
	MEDIUM_ETHERNET
} medium_t;
static medium_t current_medium;

int hgetchar(void)
{
        unsigned int c;
        static unsigned int buf;
        static unsigned int remain = 0;

        if (!remain) {
                while (!(IO_SER1_STATA & RRDY)) {
                        if (IO_SER1_STATA & SCSR_RBC) {
                                IO_SER1_STATA = SCSR_RBC;
                        }
                }
                if (IO_SER1_STATA & SCSR_RBC)
                        IO_SER1_STATA = SCSR_RBC;
                remain = IO_SER1_STATA_RXFDB;
                buf = IO_SER1_FIFO_WORD;
                remain = remain ? remain : 4;
        }
        c = buf & 0xff;
        remain -= 1;
        buf >>= 8;
        return c;
}

int hputchar(int c)
{
        if (!is_mfgi_hi()) {
                while (!(IO_SER1_STATA & TRDY));
                IO_SER1_FIFO_BYTE = (unsigned char) c;
        }
	return 0;
}

int hgetblock(unsigned char *dst, int count)
{
	int remain = count;
	if (current_medium == MEDIUM_SERIAL) {
		while (remain--)
			*dst++ = hgetchar();
	}
	return count;
}

/*
 * This works for now, since medium only affects downstream data, not
 * replies from the target to the host.  Once we add upstream media
 * switching, we'll have to implement a delayed media switch that
 * allows the successful command return value to go out along the old
 * medium.
 */
static int medium_cmdfunc(int argc, char *argv[])
{
	medium_t medium;
	
	if (argc != 2)
		return -H_EUSAGE;
	++argv;
	if (!strcmp(*argv, "ethernet"))
		medium = MEDIUM_ETHERNET;
	else if (!strcmp(*argv, "serial"))
		medium = MEDIUM_SERIAL;
	else
		return -H_EINVAL;
	if (medium == current_medium)
		return 0;
	if (medium == MEDIUM_ETHERNET) {
                return -H_EIO;
	} else if (medium == MEDIUM_SERIAL) {
//		IO_LEDFLSH = 0x7c;	/* LED continuous on */
	}
	current_medium = medium;
	return 0;
}

const command_t medium_command =
	{ "medium", "<medium name>",
	  "select communication medium",
	  &medium_cmdfunc };

